@@ -49,10 +49,10 @@ pub trait Pull {
     fn set_floating(&mut self) -> Result<&mut Self, Self::Error>;
 }
 
-pub trait Pin {
+pub trait Pin: Send {
     type Error;
 
-    fn pin() -> i32;
+    fn pin(&self) -> i32;
 
     fn reset(&mut self) -> Result<(), Self::Error>;
 }
@@ -90,6 +90,49 @@ pub struct Disabled;
 
 pub struct Unknown;
 
+// /// Generic $GpioX pin
+pub struct GpioPin<MODE> {
+    pin: i32,
+    _mode: PhantomData<MODE>,
+}
+
+impl<MODE> GpioPin<MODE>
+where
+    MODE: Send,
+{
+    fn new(pin: i32) -> GpioPin<MODE> {
+        Self {
+            pin,
+            _mode: PhantomData,
+        }
+    }
+}
+
+impl<MODE> Pin for GpioPin<MODE>
+where
+    MODE: Send,
+{
+    type Error = EspError;
+
+    fn pin(&self) -> i32
+    where
+        Self: Sized,
+    {
+        self.pin
+    }
+
+    fn reset(&mut self) -> Result<(), Self::Error> {
+        Ok(())
+    }
+}
+
+impl InputPin for GpioPin<Input> {}
+
+impl OutputPin for GpioPin<Output> {}
+
+impl OutputPin for GpioPin<InputOutput> {}
+impl InputPin for GpioPin<InputOutput> {}
+
 /// Interrupt events
 ///
 /// *Note: ESP32 has a bug (3.14), which prevents correct triggering of interrupts when
@@ -195,7 +238,20 @@ macro_rules! impl_hal_output_pin {
 
         impl embedded_hal::digital::v2::StatefulOutputPin for $pxi<$mode> {
             fn is_set_high(&self) -> Result<bool, Self::Error> {
-                Ok(unsafe { gpio_get_level($pxi::<$mode>::runtime_pin()) } != 0)
+                let pin = $pxi::<$mode>::runtime_pin() as u32;
+
+                #[cfg(esp32c3)]
+                let is_set_high = unsafe { (*(GPIO_OUT_REG as *const u32) >> pin) & 0x01 != 0 };
+                #[cfg(not(esp32c3))]
+                let is_set_high = if pin <= 31 {
+                    // GPIO0 - GPIO31
+                    unsafe { (*(GPIO_OUT_REG as *const u32) >> pin) & 0x01 != 0 }
+                } else {
+                    // GPIO32+
+                    unsafe { (*(GPIO_OUT1_REG as *const u32) >> (pin - 32)) & 0x01 != 0 }
+                };
+
+                Ok(is_set_high)
             }
 
             fn is_set_low(&self) -> Result<bool, Self::Error> {
@@ -280,22 +336,34 @@ macro_rules! impl_input_base {
         }
 
         #[cfg(not(feature = "ulp"))]
-        impl<MODE> $pxi<MODE> {
+        impl<MODE> $pxi<MODE>
+        where
+            MODE: Send,
+        {
             #[inline(always)]
             fn runtime_pin() -> i32 {
                 $pin
             }
 
+            /// Degrades a concrete pin (e.g. [`Gpio1`]) to a generic pin
+            /// struct that can also be used with periphals.
+            pub fn degrade(self) -> GpioPin<MODE> {
+                GpioPin::new($pin)
+            }
+
             pub fn into_unknown(self) -> $pxi<Unknown> {
                 $pxi { _mode: PhantomData }
             }
         }
 
-        impl<MODE> Pin for $pxi<MODE> {
+        impl<MODE> Pin for $pxi<MODE>
+        where
+            MODE: Send,
+        {
             type Error = EspError;
 
             #[inline(always)]
-            fn pin() -> i32 {
+            fn pin(&self) -> i32 {
                 $pin
             }
 
@@ -304,7 +372,7 @@ macro_rules! impl_input_base {
             }
         }
 
-        impl<MODE> InputPin for $pxi<MODE> {}
+        impl<MODE> InputPin for $pxi<MODE> where MODE: Send {}
 
         impl_hal_input_pin!($pxi: Input);
     };
@@ -315,7 +383,10 @@ macro_rules! impl_input_only {
     ($pxi:ident: $pin:expr) => {
         impl_input_base!($pxi: $pin);
 
-        impl<MODE> $pxi<MODE> {
+        impl<MODE> $pxi<MODE>
+        where
+            MODE: Send,
+        {
             /// # Safety
             ///
             /// Care should be taken not to instantiate a pin which is already used elsewhere
@@ -351,7 +422,7 @@ macro_rules! impl_input_output {
     ($pxi:ident: $pin:expr) => {
         impl_input_base!($pxi: $pin);
 
-        impl<MODE> OutputPin for $pxi<MODE> {}
+        impl<MODE> OutputPin for $pxi<MODE> where MODE: Send {}
 
         impl_hal_input_pin!($pxi: InputOutput);
         impl_hal_output_pin!($pxi: InputOutput);
@@ -359,7 +430,10 @@ macro_rules! impl_input_output {
         impl_pull!($pxi: Input);
         impl_pull!($pxi: InputOutput);
 
-        impl<MODE> $pxi<MODE> {
+        impl<MODE> $pxi<MODE>
+        where
+            MODE: Send,
+        {
             /// # Safety
             ///
             /// Care should be taken not to instantiate a pin which is already used elsewhere
@@ -442,14 +516,20 @@ macro_rules! impl_input_output {
 macro_rules! impl_rtc {
     ($pxi:ident: $pin:expr, RTC: $rtc:expr) => {
         #[cfg(feature = "ulp")]
-        impl<MODE> $pxi<MODE> {
+        impl<MODE> $pxi<MODE>
+        where
+            MODE: Send,
+        {
             #[inline(always)]
             fn runtime_pin() -> i32 {
                 $rtc
             }
         }
 
-        impl<MODE> RTCPin for $pxi<MODE> {
+        impl<MODE> RTCPin for $pxi<MODE>
+        where
+            MODE: Send,
+        {
             fn rtc_pin() -> i32 {
                 $rtc
             }
@@ -461,7 +541,10 @@ macro_rules! impl_rtc {
 
 macro_rules! impl_adc {
     ($pxi:ident: $pin:expr, ADC1: $adc:expr) => {
-        impl<MODE> ADCPin for $pxi<MODE> {
+        impl<MODE> ADCPin for $pxi<MODE>
+        where
+            MODE: Send,
+        {
             fn adc_unit() -> adc_unit_t {
                 adc_unit_t_ADC_UNIT_1
             }
@@ -473,7 +556,10 @@ macro_rules! impl_adc {
     };
 
     ($pxi:ident: $pin:expr, ADC2: $adc:expr) => {
-        impl<MODE> ADCPin for $pxi<MODE> {
+        impl<MODE> ADCPin for $pxi<MODE>
+        where
+            MODE: Send,
+        {
             fn adc_unit() -> adc_unit_t {
                 adc_unit_t_ADC_UNIT_2
             }
@@ -490,7 +576,10 @@ macro_rules! impl_adc {
 macro_rules! impl_dac {
     ($pxi:ident: $pin:expr, DAC: $dac:expr) => {
         #[cfg(all(not(esp32c3), not(esp32s3)))]
-        impl<MODE> DACPin for $pxi<MODE> {
+        impl<MODE> DACPin for $pxi<MODE>
+        where
+            MODE: Send,
+        {
             fn dac_channel() -> dac_channel_t {
                 $dac
             }
@@ -503,7 +592,10 @@ macro_rules! impl_dac {
 macro_rules! impl_touch {
     ($pxi:ident: $pin:expr, TOUCH: $touch:expr) => {
         #[cfg(not(esp32c3))]
-        impl<MODE> TouchPin for $pxi<MODE> {
+        impl<MODE> TouchPin for $pxi<MODE>
+        where
+            MODE: Send,
+        {
             fn touch_channel() -> touch_pad_t {
                 $touch
             }
